import inspect
from typing import Any, Callable, TypeVar, cast

from jararaca.messagebus import Message

DECORATED_FUNC = TypeVar("DECORATED_FUNC", bound=Callable[..., Any])
DECORATED_CLASS = TypeVar("DECORATED_CLASS", bound=Any)


MESSAGEBUS_INCOMING_MAP = dict[str, Callable[[Message[Any]], Any]]


class MessageBusController:

    MESSAGEBUS_ATTR = "__messagebus__"

    def __init__(self) -> None:
        self.messagebus_factory: Callable[[Any], MESSAGEBUS_INCOMING_MAP] | None = None

    def get_messagebus_factory(
        self,
    ) -> Callable[[DECORATED_CLASS], MESSAGEBUS_INCOMING_MAP]:
        if self.messagebus_factory is None:
            raise Exception("MessageBus factory is not set")
        return self.messagebus_factory

    def __call__(self, func: type[DECORATED_CLASS]) -> type[DECORATED_CLASS]:

        def messagebus_factory(
            instance: DECORATED_CLASS,
        ) -> MESSAGEBUS_INCOMING_MAP:
            handlers: MESSAGEBUS_INCOMING_MAP = {}
            inspect.signature(func)

            members = inspect.getmembers(func, predicate=inspect.isfunction)

            for name, member in members:
                message_incoming = IncomingHandler.get_message_incoming(member)

                if message_incoming is None:
                    continue

                if not inspect.iscoroutinefunction(member):
                    raise Exception(
                        "Message incoming handler '%s' from '%s.%s' must be a coroutine function"
                        % (name, func.__module__, func.__qualname__)
                    )

                handlers[message_incoming.topic] = getattr(instance, name)

            return handlers

        self.messagebus_factory = messagebus_factory

        MessageBusController.register(func, self)

        return func

    @staticmethod
    def register(
        func: type[DECORATED_CLASS], messagebus: "MessageBusController"
    ) -> None:

        setattr(func, MessageBusController.MESSAGEBUS_ATTR, messagebus)

    @staticmethod
    def get_messagebus(func: type[DECORATED_CLASS]) -> "MessageBusController | None":
        if not hasattr(func, MessageBusController.MESSAGEBUS_ATTR):
            return None

        return cast(
            MessageBusController, getattr(func, MessageBusController.MESSAGEBUS_ATTR)
        )


class IncomingHandler:

    MESSAGE_INCOMING_ATTR = "__message_incoming__"

    def __init__(
        self,
        topic: str,
        timeout: int | None = None,
        exception_handler: Callable[[BaseException], None] | None = None,
        nack_on_exception: bool = False,
        auto_ack: bool = True,
    ) -> None:
        self.topic = topic
        self.timeout = timeout
        self.exception_handler = exception_handler
        self.nack_on_exception = nack_on_exception
        self.auto_ack = auto_ack

    def __call__(self, func: DECORATED_FUNC) -> DECORATED_FUNC:

        IncomingHandler.register(func, self)

        return func

    @staticmethod
    def register(func: DECORATED_FUNC, message_incoming: "IncomingHandler") -> None:

        setattr(func, IncomingHandler.MESSAGE_INCOMING_ATTR, message_incoming)

    @staticmethod
    def get_message_incoming(func: DECORATED_FUNC) -> "IncomingHandler | None":
        if not hasattr(func, IncomingHandler.MESSAGE_INCOMING_ATTR):
            return None

        return cast(
            IncomingHandler, getattr(func, IncomingHandler.MESSAGE_INCOMING_ATTR)
        )



import inspect
from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Literal, Protocol, Type, TypeVar, cast
from urllib.parse import urljoin

import httpx
from pydantic import BaseModel

DECORATED_FUNC = TypeVar("DECORATED_FUNC", bound=Callable[..., Awaitable[Any]])


class HttpMapping:

    HTTP_MAPPING_ATTR = "__rest_http_client_mapping__"

    def __init__(self, method: str, path: str, success_status: int = 200):
        self.method = method
        self.path = path
        self.success_status = success_status

    def __call__(self, func: DECORATED_FUNC) -> DECORATED_FUNC:
        HttpMapping.register(func, self)
        return func

    @staticmethod
    def register(funf: DECORATED_FUNC, instance: "HttpMapping") -> None:
        setattr(funf, HttpMapping.HTTP_MAPPING_ATTR, instance)

    @staticmethod
    def get(funf: DECORATED_FUNC) -> "HttpMapping | None":
        if hasattr(funf, HttpMapping.HTTP_MAPPING_ATTR):
            return cast(HttpMapping, getattr(funf, HttpMapping.HTTP_MAPPING_ATTR))

        return None


class Post(HttpMapping):

    def __init__(self, path: str):
        super().__init__("POST", path)


class Get(HttpMapping):

    def __init__(self, path: str):
        super().__init__("GET", path)


class Patch(HttpMapping):

    def __init__(self, path: str):
        super().__init__("PATCH", path)


class Put(HttpMapping):

    def __init__(self, path: str):
        super().__init__("PUT", path)


class Delete(HttpMapping):

    def __init__(self, path: str):
        super().__init__("DELETE", path)


class RequestAttribute:

    REQUEST_ATTRIBUTE_ATTRS = "__request_attributes__"

    @staticmethod
    def register(cls: DECORATED_FUNC, instance: "RequestAttribute") -> None:

        if not hasattr(cls, RequestAttribute.REQUEST_ATTRIBUTE_ATTRS):
            setattr(cls, RequestAttribute.REQUEST_ATTRIBUTE_ATTRS, [])

        getattr(cls, RequestAttribute.REQUEST_ATTRIBUTE_ATTRS).append(instance)

    @staticmethod
    def get(cls: DECORATED_FUNC) -> "list[RequestAttribute]":
        if hasattr(cls, RequestAttribute.REQUEST_ATTRIBUTE_ATTRS):
            return cast(
                list[RequestAttribute],
                getattr(cls, RequestAttribute.REQUEST_ATTRIBUTE_ATTRS),
            )

        return []

    def __init__(
        self, attribute_type: Literal["query", "header", "body", "param"], name: str
    ):
        self.attribute_type = attribute_type
        self.name = name

    def __call__(self, cls: DECORATED_FUNC) -> DECORATED_FUNC:
        RequestAttribute.register(cls, self)
        return cls


class Query(RequestAttribute):

    def __init__(self, name: str):
        super().__init__("query", name)


class Header(RequestAttribute):

    def __init__(self, name: str):
        super().__init__("header", name)


class Body(RequestAttribute):

    def __init__(self, name: str):
        super().__init__("body", name)


class PathParam(RequestAttribute):

    def __init__(self, name: str):
        super().__init__("param", name)


DECORATED_CLASS = TypeVar("DECORATED_CLASS", bound=Any)


class RestClient:

    REST_CLIENT_ATTR = "__rest_client__"

    def __init__(self, base_path: str) -> None:
        self.base_path = base_path

    @staticmethod
    def register(cls: type, instance: "RestClient") -> None:
        setattr(cls, RestClient.REST_CLIENT_ATTR, instance)

    @staticmethod
    def get(cls: type) -> "RestClient | None":
        if hasattr(cls, RestClient.REST_CLIENT_ATTR):
            return cast(RestClient, getattr(cls, RestClient.REST_CLIENT_ATTR))

        return None

    def __call__(self, cls: Type[DECORATED_CLASS]) -> Type[DECORATED_CLASS]:

        RestClient.register(cls, self)

        return cls


class RPCRequestNetworkError(Exception):
    pass


@dataclass
class HttpRPCResponse:

    status_code: int
    data: bytes


@dataclass
class HttpRPCRequest:
    url: str
    method: str
    headers: list[tuple[str, str]]
    query_params: dict[str, str]
    body: bytes | None


class HttpRPCAsyncBackend(Protocol):

    async def request(
        self,
        request: HttpRPCRequest,
    ) -> HttpRPCResponse: ...


class HTTPXHttpRPCAsyncBackend(HttpRPCAsyncBackend):

    def __init__(self, prefix_url: str = ""):
        self.prefix_url = prefix_url

    async def request(
        self,
        request: HttpRPCRequest,
    ) -> HttpRPCResponse:

        async with httpx.AsyncClient() as client:

            try:
                response = await client.request(
                    method=request.method,
                    url=urljoin(self.prefix_url, request.url),
                    headers=request.headers,
                    params=request.query_params,
                    content=request.body,
                )

                return HttpRPCResponse(
                    status_code=response.status_code,
                    data=response.content,
                )
            except httpx.NetworkError:
                raise RPCRequestNetworkError("Network error")


T = TypeVar("T")


class RequestMiddleware(Protocol):

    def on_request(self, request: HttpRPCRequest) -> HttpRPCRequest: ...


from opentelemetry import baggage, trace
from opentelemetry.baggage.propagation import W3CBaggagePropagator
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator


class TracedRequestMiddleware(RequestMiddleware):

    def on_request(self, request: HttpRPCRequest) -> HttpRPCRequest:

        span = trace.get_current_span()

        if span is not None:
            ctx = baggage.set_baggage("hello", "world")
            headers: dict[str, str] = {}
            W3CBaggagePropagator().inject(headers, ctx)
            TraceContextTextMapPropagator().inject(headers, ctx)

            for key, value in headers.items():
                request.headers.append((key, value))

        return request


class HttpRpcClientBuilder:

    def __init__(
        self,
        backend: HttpRPCAsyncBackend,
        middlewares: list[RequestMiddleware] = [],
    ):
        self._backend = backend
        self._middlewares = middlewares

    def build(self, cls: type[T]) -> T:
        rest_client = RestClient.get(cls)

        if rest_client is None:
            raise ValueError("Class is not a rest client")

        def create_method(
            mapping: HttpMapping,
            method_call: Callable[..., Any],
            http_method: str,
            path: str,
        ) -> Callable[..., Awaitable[Any]]:

            call_signature = inspect.signature(method_call)
            call_parameters = [*call_signature.parameters.keys()][1:]

            async def rpc_method(*args: Any, **kwargs: Any) -> Any:

                args_as_kwargs = dict(zip(call_parameters, args))

                request_attributes = RequestAttribute.get(method_call)

                compiled_kwargs = {**args_as_kwargs, **kwargs}

                headers: list[tuple[str, str]] = []
                query_params = {}
                body: Any = None
                compiled_path = (
                    rest_client.base_path.rstrip("/") + "/" + mapping.path.lstrip("/")
                )

                for attr in request_attributes:
                    if attr.attribute_type == "header":
                        headers.append((attr.name, compiled_kwargs[attr.name]))
                    elif attr.attribute_type == "query":
                        query_params[attr.name] = compiled_kwargs[attr.name]
                    elif attr.attribute_type == "body":
                        body = compiled_kwargs[attr.name]
                    elif attr.attribute_type == "param":
                        compiled_path = path.replace(
                            f":{attr.name}", str(compiled_kwargs[attr.name])
                        )

                body_content: bytes | None = None

                if body is not None:
                    if isinstance(body, BaseModel):
                        body_content = body.model_dump_json().encode()
                        headers.append(("Content-Type", "application/json"))
                    elif isinstance(body, bytes):
                        body_content = body
                    else:
                        raise ValueError("Invalid body type")

                request = HttpRPCRequest(
                    url=compiled_path,
                    method=http_method,
                    headers=headers,
                    query_params=query_params,
                    body=body_content,
                )

                for middleware in self._middlewares:
                    request = middleware.on_request(request)

                response = await self._backend.request(request)

                return_type = inspect.signature(method_call).return_annotation

                if response.status_code != mapping.success_status:
                    raise ValueError(
                        "Invalid status code: {}".format(response.status_code)
                    )

                if return_type is not inspect.Signature.empty:
                    if issubclass(return_type, BaseModel):
                        return return_type.model_validate_json(response.data)
                    if return_type is bytes:
                        return response.data
                    if return_type is None:
                        return None

                return response

            return rpc_method

        class Dummy: ...

        dummy = Dummy()

        for attr_name in dir(cls):
            method_call = getattr(cls, attr_name)
            if (mapping := HttpMapping.get(method_call)) is not None:
                setattr(
                    dummy,
                    attr_name,
                    create_method(mapping, method_call, mapping.method, mapping.path),
                )

        return cast(T, dummy)



import logging
from contextlib import asynccontextmanager, contextmanager
from typing import AsyncContextManager, AsyncGenerator, Generator, Protocol

from opentelemetry import metrics, trace
from opentelemetry._logs import set_logger_provider
from opentelemetry.baggage.propagation import W3CBaggagePropagator
from opentelemetry.exporter.otlp.proto.http._log_exporter import (
    OTLPLogExporter as LogExporter,
)
from opentelemetry.exporter.otlp.proto.http.metric_exporter import (
    OTLPMetricExporter as MeterExporter,
)
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as SpanExporter,
)
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

from jararaca.microservice import (
    AppContext,
    AppInterceptor,
    AppInterceptorWithLifecycle,
    Container,
    Microservice,
)
from jararaca.observability.decorators import (
    TracingContextProvider,
    TracingContextProviderFactory,
    provide_tracing_ctx_provider,
)

tracer = trace.get_tracer(__name__)


class OtelTracingContextProvider(TracingContextProvider):

    def __init__(self, app_context: AppContext) -> None:
        self.app_context = app_context

    @contextmanager
    def __call__(
        self,
        trace_name: str,
        context_attributes: dict[str, str],
    ) -> Generator[None, None, None]:

        with tracer.start_as_current_span(trace_name, attributes=context_attributes):
            yield


class OtelTracingContextProviderFactory(TracingContextProviderFactory):

    def provide_provider(self, app_context: AppContext) -> TracingContextProvider:
        return OtelTracingContextProvider(app_context)

    @asynccontextmanager
    async def root_setup(self, app_context: AppContext) -> AsyncGenerator[None, None]:

        title: str = "Unmapped App Context Execution"
        headers = {}

        if app_context.context_type == "http":
            print(
                "##############Span is not recording for url %s"
                % app_context.request.url
            )
            headers = dict(app_context.request.headers)
            title = f"HTTP {app_context.request.method} {app_context.request.url}"

        elif app_context.context_type == "message_bus":
            title = f"Message Bus {app_context.topic}"

        carrier = {
            key: value
            for key, value in headers.items()
            if key.lower().startswith("traceparent")
            or key.lower().startswith("tracestate")
        }

        ctx = TraceContextTextMapPropagator().extract(carrier)

        b2 = {
            key: value
            for key, value in headers.items()
            if key.lower().startswith("baggage")
        }

        ctx2 = W3CBaggagePropagator().extract(b2, context=ctx)

        with tracer.start_as_current_span(name=title, context=ctx2):
            yield


class ObservabilityProvider(Protocol):

    tracing_provider: TracingContextProviderFactory

    def setup(
        self, app: Microservice, container: Container
    ) -> AsyncContextManager[None]: ...


class OtelObservabilityProvider(ObservabilityProvider):

    def __init__(
        self,
        app_name: str,
        logs_exporter: LogExporter,
        span_exporter: SpanExporter,
        meter_exporter: MeterExporter,
    ) -> None:
        self.app_name = app_name
        self.logs_exporter = logs_exporter
        self.span_exporter = span_exporter
        self.meter_exporter = meter_exporter
        self.tracing_provider = OtelTracingContextProviderFactory()

    @asynccontextmanager
    async def setup(
        self, app: Microservice, container: Container
    ) -> AsyncGenerator[None, None]:
        ### Setup Resource

        resource = Resource(attributes={SERVICE_NAME: self.app_name})

        ### Setup Tracing
        provider = TracerProvider(resource=resource)

        trace.set_tracer_provider(provider)

        span_processor = BatchSpanProcessor(self.span_exporter)
        provider.add_span_processor(span_processor)

        ### Setup Logs
        logger_provider = LoggerProvider(resource=resource)

        set_logger_provider(logger_provider)

        logger_provider.add_log_record_processor(
            BatchLogRecordProcessor(self.logs_exporter)
        )

        logging_handler = LoggingHandler(
            level=logging.DEBUG, logger_provider=logger_provider
        )
        logging_handler.addFilter(
            lambda record: hasattr(record, "message")
            and "localhost" not in record.message
        )
        logging.getLogger().addHandler(logging_handler)
        logging.getLogger().setLevel(logging.DEBUG)

        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )

        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        logging.getLogger().addHandler(stream_handler)

        ### Setup Metrics
        metric_reader = PeriodicExportingMetricReader(
            self.meter_exporter, export_interval_millis=1000
        )
        meter_provider = MeterProvider(metric_readers=[metric_reader])

        metrics.set_meter_provider(meter_provider)

        yield

    @staticmethod
    def from_url(app_name: str, url: str) -> "OtelObservabilityProvider":
        """
        Create an instance of OtelObservabilityProvider with Http Exporters from a given URL
        """

        logs_exporter = LogExporter(endpoint=f"{url}/v1/logs")
        span_exporter = SpanExporter(endpoint=f"{url}/v1/traces")
        metric_exporter = MeterExporter(endpoint=f"{url}/v1/metrics")

        return OtelObservabilityProvider(
            app_name, logs_exporter, span_exporter, metric_exporter
        )


class ObservabilityInterceptor(AppInterceptor, AppInterceptorWithLifecycle):

    def __init__(
        self,
        observability_provider: ObservabilityProvider,
    ) -> None:
        self.observability_provider = observability_provider

    @asynccontextmanager
    async def intercept(self, app_context: AppContext) -> AsyncGenerator[None, None]:

        async with self.observability_provider.tracing_provider.root_setup(app_context):

            with provide_tracing_ctx_provider(
                self.observability_provider.tracing_provider.provide_provider(
                    app_context
                )
            ):
                yield

    @asynccontextmanager
    async def lifecycle(
        self, app: Microservice, container: Container
    ) -> AsyncGenerator[None, None]:

        # resource = Resource(attributes={SERVICE_NAME: "jararaca-example"})
        # span_exporter = SpanExporter(endpoint="http://localhost:4318/v1/traces")

        # provider = TracerProvider(resource=resource)
        # trace.set_tracer_provider(provider)

        # span_processor = BatchSpanProcessor(span_exporter)

        # provider.add_span_processor(span_processor)

        # ####

        # logs_exporter = LogExporter(endpoint="http://localhost:4318/v1/logs")
        # set_logger_provider(logger_provider := LoggerProvider(resource=resource))
        # logger_provider.add_log_record_processor(BatchLogRecordProcessor(logs_exporter))

        # logging_handler = LoggingHandler(
        #     level=logging.DEBUG, logger_provider=logger_provider
        # )

        # logging.getLogger().addHandler(logging_handler)
        # logging.getLogger().addHandler(logging.StreamHandler())
        # logging.getLogger().setLevel(logging.DEBUG)
        # logging.getLogger().info("Logging initialized")
        async with self.observability_provider.setup(app, container):
            yield



from typing import Generic, Protocol, TypeVar

T = TypeVar("T", covariant=True)


class Message(Protocol, Generic[T]):

    def payload(self) -> T: ...

    async def ack(self) -> None: ...

    async def reject(self) -> None: ...

    async def nack(self) -> None: ...
